// File for Belle2 configuration with PCIe40 
#include "pcie40_b2config.h"
#include <string.h>
#include <map>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <fstream>
#include <vector>
#include <unistd.h>

// ===== 
// CALL C functions
// =====
extern "C" int ecs_iordLli(int dev, unsigned add, unsigned *val);
extern "C" int mpod_channelDisable(int dev, int mpid, int channel, int flag);
extern "C" int mpod_read(int dev, int mpid, char reg, unsigned *val) ;
extern "C" unsigned ecs_read(int dev, int bar, unsigned add); 
// ===
// Mapping of minipod
// === 
int mpodid[ 48 ] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 
		     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 
		     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7 } ;

int channelid[ 48 ] = { 6, 10, 8, 4, 2 , 0, 1, 3, 5, 7, 9, 11, 
			10, 8, 6, 4, 2, 0, 1, 5, 3, 7, 9, 11,
			10, 8, 6, 4, 2, 0, 1, 3, 5, 7, 9, 11,
			0, 2, 4, 6, 8, 10, 1, 3, 5, 7, 9, 11 } ;

std::string pcie40_fpgaVersion( int dev ) {
  int reg = 0 ;
  unsigned val ; 
  std::string content ;
  for ( int i = 0 ; i < SIZE_MEMO_FPGA ; ++i ) {
    ecs_iordLli( dev , reg , &val ) ;
    content += val ;
    reg += 4 ; 
  }

  std::stringstream ss( content ) ;
  std::string to ;
  bool found = false ;
  while ( std::getline( ss , to , '\n' ) ) {
    std::stringstream sss( to ) ;
    std::string tok ;
    while ( std::getline( sss , tok , ':' ) ) {
      if ( found ) {
	tok.erase(std::remove(tok.begin(), tok.end(), ' '), 
		  tok.end());
	tok.erase(std::remove(tok.begin(), tok.end(), '\"'), 
		  tok.end());
	tok.erase(std::remove(tok.begin(), tok.end(), ','), 
		  tok.end());
	return tok ;
      }
      if ( tok.rfind( "\"design_version\"" ) == 0 ) found = true ;
    }
  }
  return "unknown" ; 
}

/// Disable one TX link (physically via the minipod configuration)
void pcie40_disableMinipodLink( int dev , int ch ) {
  int mpod = mpodid[ ch ] ;
  int channel = channelid[ ch ] ;
  int result = mpod_channelDisable( dev , mpod , channel , 1 ) ;
  if ( 0 != result ) std::cout << "Error disable minipod" << std::endl ;
}


/// Enable one TX link (physically via the minipod configuration)
void pcie40_enableMinipodLink( int dev , int ch ) {
  int mpod = mpodid[ ch ] ;
  int channel = channelid[ ch ] ;
  int result = mpod_channelDisable( dev , mpod , channel , 0 ) ;
  if ( 0 != result ) std::cout << "Error enable minipod" << std::endl ;
}

/// Reset one optical link via minipod
void pcie40_resetMinipodLink(int dev, int ch ) {
  pcie40_disableMinipodLink( dev , ch ) ; 
  sleep( 1 ) ; 
  pcie40_enableMinipodLink( dev , ch ) ; 
}

/// Gives the status (on = 1, off = 0) of the minipod link
int pcie40_statusMinipodLink(int dev, int ch ) {
  int mpid = mpodid[ ch ] ;
  unsigned val ; 
  int total ; 
  int result = mpod_read(dev, mpid, 93 , &val ) ;
  if ( 0 != result ) 
    std::cout << "Error dumping channel status" << std::endl ;
  total = val & 0xFF ; 

  result = mpod_read(dev, mpid, 92 , &val ) ;
  if ( 0 != result ) 
    std::cout << "Error dumping channel status" << std::endl ;
  total = total | ( ( val & 0xF ) << 8 ) ;
  if ( ( total & ( 1 << ( channelid[ ch ] ) ) ) == 1 ) return 0 ;
  return 1 ;
}

////////////////////////////////////////////////////////////////////////////////////
// Internal functions for PLLs
//////////////////////////////////////////////////////////////////////////////////// 
extern "C" int i2c_init(int dev, int bus, unsigned inCLK, unsigned SCLfreq);
extern "C" int i2c_readMem(int dev, int bus, unsigned slaveAdd, unsigned regIndex, unsigned *val);
extern "C" int i2c_writeMem(int dev, int bus, unsigned slaveAdd, unsigned regIndex, unsigned *val);

class pcie40_pll { 
public:
  pcie40_pll( int device , int bus_number , int address ) { 
    PAGE = 0x0001 ;
    dev = device ;
    bus = bus_number ; 
    add = address ; 
    initDone = false ; 
  }
  void hardReset( ) ;
  void programming( std::ifstream & file ) ;
  void enable_los( int ch ) ;
  void enable_oof( int ch ) ;
  void clear_los_oof_flg_reg();  
  void clear_lol_flg_bit();
  bool read_loss_of_lock() ;
  bool read_loss_of_signal( int ch ) ;
  bool read_out_of_frequency( int ch ) ;
  bool read_loss_of_lock_flg() ;
  bool read_loss_of_signal_flg( int ch ) ;
  bool read_out_of_frequency_flg( int ch ) ;

private:
  int read_pagined_register( int reg , unsigned int & val ) ;
  int write_pagined_register( int reg , unsigned int val ) ;
  void init_i2c() ; 
  void set_bit( int reg , int offset ) ;
  void clear_bit( int reg , int offset ) ;
  int PAGE ; 
  int dev ;
  int bus ; 
  int add ; 
  bool initDone;
};

void pcie40_pll::set_bit( int reg , int offset ) {
  unsigned int val ; 
  int status = read_pagined_register(reg, val ) ;
  unsigned int mask = 1 << offset ;
  status = write_pagined_register(reg, val | mask) ;
}  

void pcie40_pll::clear_bit( int reg , int offset ) {
  unsigned int val ;
  int status = read_pagined_register(reg, val) ;
  unsigned int mask = ~(1 << offset) ;
  status = write_pagined_register(reg, val & mask);
}

void pcie40_pll::init_i2c() { 
  i2c_init(dev,bus, 100000000, 100000) ;
  initDone = true ; 
}

int pcie40_pll::read_pagined_register( int register_address , unsigned int & val ) {
  if ( ! initDone ) init_i2c() ;
  unsigned int msbyte_register_address = register_address >> 8 ;
  int lsbyte_register_address = 0xFF & register_address ;
  int page_register_address = PAGE ;
  unsigned int current_page ; 
  int status1 = i2c_readMem( dev, bus, add, page_register_address, &current_page ) ;
  current_page = current_page & 0xFFFFFFFF ;
  if ( current_page == 0xFFFFFFFF ) {
    std::cout << "Problem in reading PLL register" << std::endl ;
  }

  int status2 ;
  if ( current_page != msbyte_register_address ) {
    status2 = i2c_writeMem( dev, bus, add, page_register_address, &msbyte_register_address ) ;
  } else {
    status2 = 0;
  }
  int status3 = i2c_readMem(dev, bus, add, lsbyte_register_address, &val );
  val = val & 0xFFFFFFFF ;
  int status = 0 ;
  if ( ( status1 == -1 ) || ( status2 == -1 ) || ( status3 == -1 ) )
    status = -1 ;
  return status ;
}

int pcie40_pll::write_pagined_register( int register_address , unsigned int value_to_write ) {
  if ( ! initDone ) init_i2c() ;
  unsigned int msbyte_register_address = register_address >> 8 ;
  int lsbyte_register_address = 0xFF & register_address;
  int page_register_address = PAGE ;
  unsigned int current_page ; 
  int status1 = i2c_readMem(dev, bus, add, page_register_address, &current_page ) ;
  int status2 ; 
  current_page = current_page & 0xFFFFFFFF ; 
  if ( current_page != msbyte_register_address ) {
    status2 = i2c_writeMem(dev, bus, add, page_register_address, &msbyte_register_address ) ;
  } else { 
    status2 = 0 ;
  }
  int status3 = i2c_writeMem(dev, bus, add, lsbyte_register_address, &value_to_write ) ;
  int status = 0; 
  if ( ( status1 == -1 ) || ( status2 == -1 ) || ( status3 == -1 ) ) 
    status = -1;
  return status;
}

void pcie40_pll::hardReset( ) {
  set_bit(0x001E, 0);
  clear_bit(0x001E, 0);
  set_bit(0x001E, 1);
  clear_bit(0x001E, 1);
}

void pcie40_pll::programming( std::ifstream & file ) {
  // **********************************1) in first preamble***********************************
  //header suppress
  int indice_entete = 0;
  std::string line ; 
  while ( std::getline( file , line ) ) {
    if ( line.find( "Start configuration preamble") != std::string::npos ) break ;
    indice_entete += 1 ;
  }

  while ( std::getline( file , line ) ) {
    if ( line.find( "End configuration preamble" ) != std::string::npos ) break ;
    std::istringstream ss( line ) ;
    std::string s ; 
    std::vector< std::string > strings ;
    while( std::getline( ss , s , ',' ) ) strings.push_back( s ) ; 
    int current_regstr = std::strtol( strings[ 0 ].c_str() , NULL , 16 )  ; 
    unsigned int current_valstr = std::strtoul( strings[ 1 ].c_str() , NULL , 16 ) ;
    // int status1 = 
    write_pagined_register(current_regstr, current_valstr) ;
  }
  // **********************************2) wait 300ms *****************************************
  sleep(0.3) ;
  // ********************3) Then perform the desired register modifications ******************
  indice_entete = 0;
  while ( std::getline( file , line ) ) { 
    if ( line.find("Start configuration registers") != std::string::npos ) break ; 
    indice_entete += 1 ;
  }

  while ( std::getline( file , line ) ) {
    if ( line.find( "End configuration registers" ) != std::string::npos ) break ;
    std::istringstream ss( line ) ;
    std::string s ;
    std::vector< std::string > strings ;
    while( std::getline( ss , s , ',' ) ) strings.push_back( s ) ;
    int current_regstr = std::strtol( strings[ 0 ].c_str() , NULL , 16 )  ;
    unsigned int current_valstr = std::strtoul( strings[ 1 ].c_str() , NULL , 16 ) ;
    // int status1 = 
    write_pagined_register(current_regstr, current_valstr) ;
  }    

  // ********************4) Write the post-amble and SOFT_RST ********************************
  // text suppress
  indice_entete = 0;
  while ( std::getline( file , line ) ) {
    if ( line.find("Start configuration postamble") != std::string::npos ) break ;
    indice_entete += 1 ;
  }
  while ( std::getline( file , line ) ) {
    if ( line.find( "End configuration postamble" ) != std::string::npos ) break ;
    std::istringstream ss( line ) ;
    std::string s ;
    std::vector< std::string > strings ;
    while( std::getline( ss , s , ',' ) ) strings.push_back( s ) ;
    int current_regstr = std::strtol( strings[ 0 ].c_str() , NULL , 16 )  ;
    unsigned int current_valstr = std::strtoul( strings[ 1 ].c_str() , NULL , 16 ) ;
    //int status1 = 
    write_pagined_register(current_regstr, current_valstr) ;
  }
}

void pcie40_pll::enable_los( int ch ) { 
  set_bit( 0x002C , ch ) ;
}

void pcie40_pll::enable_oof( int ch ) { 
  set_bit( 0x003F , ch );
}

void pcie40_pll::clear_los_oof_flg_reg() { 
  unsigned int current_alarms , enabled_alarms; 
  int status = read_pagined_register(0x000D, current_alarms ) ;
  status = read_pagined_register(0x002C, enabled_alarms ) ;
  status = write_pagined_register(0x002C, enabled_alarms & ~(current_alarms & 0xF));
  status = write_pagined_register(0x0012, 0x00);
  status = write_pagined_register(0x002C, enabled_alarms | (current_alarms & 0xF)) ;
}

void pcie40_pll::clear_lol_flg_bit() { 
  int register_address = 0x0013 ;
  unsigned int value ;
  int status1 = read_pagined_register(register_address, value ) ;
  value = value & ~(1 << 1) ;
  // int status2 = 
  write_pagined_register(register_address, value) ;
  status1 = read_pagined_register(register_address , value ) ;
}

bool pcie40_pll::read_loss_of_lock() { 
  int register_address = 0x000E ;
  int bit_position = 1 ;
  bool loss_of_lock = false ;
  unsigned int value ;
  // int status = 
  read_pagined_register(register_address, value) ;
  if ( (value >> bit_position) & 1 ) loss_of_lock = true ;
  return loss_of_lock ;
}

bool pcie40_pll::read_loss_of_lock_flg() { 
  int register_address = 0x0013 ;
  int bit_position = 1 ;
  bool loss_of_lock_flg = false ;
  unsigned int value ;
  // int status = 
  read_pagined_register(register_address, value) ;
  if ( (value >> bit_position) & 1 ) loss_of_lock_flg = true ;
  return loss_of_lock_flg ;
}

bool pcie40_pll::read_loss_of_signal( int ch ) { 
  if ( ch > 3 ) {
    ch = 3 ; 
    std::cout << "PLL read LOS: wrong number" << std::endl ;
  }
  int register_address = 0x000D ;
  int bit_position = ch ;
  bool loss_of_signal = false ;
  unsigned int value ;
  // int status = 
  read_pagined_register(register_address, value) ;
  if ( (value >> bit_position) & 1 ) loss_of_signal = true ;
  return loss_of_signal ;
}

bool pcie40_pll::read_loss_of_signal_flg( int ch ) { 
  if ( ch > 3 ) {
    ch = 3 ; 
    std::cout << "PLL read LOS flag: wrong number" << std::endl ;
  }
  int register_address = 0x0012 ;
  int bit_position = ch ;
  bool loss_of_signal_flg = false ;
  unsigned int value ;
  // int status = 
  read_pagined_register(register_address, value) ;
  if ( (value >> bit_position) & 1 ) loss_of_signal_flg = true ;
  return loss_of_signal_flg ;
}

bool pcie40_pll::read_out_of_frequency( int ch ) { 
  if ( ch > 3 ) {
    ch = 3 ; 
    std::cout << "PLL read OOF: wrong number" << std::endl ;
  }
  int register_address = 0x000D ;
  int bit_position = 4 + ch ;
  bool out_of_frequency = false ;
  unsigned int value ;
  // int status = 
  read_pagined_register(register_address, value) ;
  if ( (value >> bit_position) & 1 ) out_of_frequency = true ;
  return out_of_frequency ;
}

bool pcie40_pll::read_out_of_frequency_flg( int ch ) { 
  if ( ch > 3 ) {
    ch = 3 ; 
    std::cout << "PLL read OOF: wrong number" << std::endl ;
  }
  int register_address = 0x0012 ;
  int bit_position = 4 + ch ;
  bool out_of_frequency_flg = false ;
  unsigned int value ;
  //int status = 
  read_pagined_register(register_address, value) ;
  if ( (value >> bit_position) & 1 ) out_of_frequency_flg = true ;
  return out_of_frequency_flg ;
}

/// Configure the PLLs for belle2 clock usage
int pcie40_configurePLLs( int dev ) {
  pcie40_pll pll1( dev , 1 , 0x68 ) ;
  pcie40_pll pll2( dev , 2 , 0x68 ) ; 
  pcie40_pll pll3( dev , 3 , 0x68 ) ; 

  // 1. reset the PLLs ( pll_si5345_1, pll_si5345_2, pll_si5344 ) 
  // int status = 0 ; 
  pll1.hardReset() ; 
  pll2.hardReset() ; 
  pll3.hardReset() ;

  // 2. program the 3 PLLs from files:
  // 1: "$PLLDATAPATH/Si5345-1-BII_FTSW-Registers.txt",
  //  2: "$PLLDATAPATH/Si5345-2-BII_FTSW-Registers.txt",
  //  3: "$PLLDATAPATH/Si5344-BII_FTSW-Registers.txt"
  // Files must be in directory from env variable "PLLDATAPATH" 
  if ( std::getenv( "PLLDATAPATH" ) == NULL ) {
    std::cout << "PLLDATAPATH variable not defined" << std::endl ;
    return -1 ;
  }
  std::string data_path( std::getenv( "PLLDATAPATH" ) ) ;
  std::string filename1 = data_path + std::string( "/Si5345-1-BII_FTSW-Registers.txt" ) ;
  std::ifstream data1( filename1.c_str() ) ;
  if ( ! data1.good() ) { 
    std::cout << "Data file1 for PLL does not exist " << filename1 << std::endl ;
    return -1 ; 
  }
  pll1.programming( data1 ) ; 
  data1.close() ;

  std::string filename2 = data_path + std::string( "/Si5345-2-BII_FTSW-Registers.txt" ) ;
  std::ifstream data2( filename2.c_str() ) ;
  if ( ! data2.good() ) { 
    std::cout << "Data file2 for PLL does not exist " << filename2 << std::endl ;
    return -1 ; 
  }
  pll2.programming( data2 ) ; 
  data2.close() ;

  std::string filename3 = data_path + std::string( "/Si5344-BII_FTSW-Registers.txt" ) ;
  std::ifstream data3( filename3.c_str() ) ;
  if ( ! data3.good() ) { 
    std::cout << "Data file3 for PLL does not exist " << filename3 << std::endl ;
    return -1 ; 
  }
  pll3.programming( data3 ) ; 
  data3.close() ;

  // 3. enable LOS and OOF monitoring
  pll1.enable_los( 0 ) ; 
  pll1.enable_los( 1 ) ; 
  pll1.enable_los( 2 ) ; 
  pll1.enable_los( 3 ) ; 

  pll2.enable_los( 0 ) ; 
  pll2.enable_los( 1 ) ; 
  pll2.enable_los( 2 ) ; 
  pll2.enable_los( 3 ) ; 

  pll3.enable_los( 0 ) ; 
  pll3.enable_los( 1 ) ; 
  pll3.enable_los( 2 ) ; 
  pll3.enable_los( 3 ) ; 

  pll1.enable_oof( 0 ) ; 
  pll1.enable_oof( 1 ) ; 
  pll1.enable_oof( 2 ) ; 
  pll1.enable_oof( 3 ) ; 

  pll2.enable_oof( 0 ) ; 
  pll2.enable_oof( 1 ) ; 
  pll2.enable_oof( 2 ) ; 
  pll2.enable_oof( 3 ) ; 

  pll3.enable_oof( 0 ) ; 
  pll3.enable_oof( 1 ) ; 
  pll3.enable_oof( 2 ) ; 
  pll3.enable_oof( 3 ) ; 

  // 4. clear LOS and OOF counters
  sleep( 7 ) ;
  pll1.clear_lol_flg_bit() ;
  pll1.clear_los_oof_flg_reg() ;
  pll2.clear_lol_flg_bit() ;
  pll2.clear_los_oof_flg_reg() ;
  pll3.clear_lol_flg_bit() ;
  pll3.clear_los_oof_flg_reg() ;

  return 0 ;
}

int pcie40_inputPLLFrequency( int dev ) { 
  return ecs_read( dev , 2 , 0x1140 ) ; 
}

/// Status of PLL (locked = 1 , not locked = 0 ), for the 3 external PLLs 
/// ( 1 = PLL for input channels 0 to 23, 2 = PLL for input channels 24 to 47, 3 = FTSW PLL)
int pcie40_pllLockStatus( int dev , int pll ) {
  if ( pll > 3 ) {
    std::cout << "Invalid PLL number" << std::endl ;
    return 0 ;
  }
  pcie40_pll the_pll( dev , pll , 0x68 ) ;
  if ( the_pll.read_loss_of_lock() == true ) return 0 ;
  return 1 ;
}

/// Input status of PLL (good = 1 , loss of signal = 0)
int pcie40_pllInputStatus( int dev , int pll ) {
  if ( pll > 3 ) {
    std::cout << "Invalid PLL number" << std::endl ;
    return 0 ;
  }
  pcie40_pll the_pll( dev , pll , 0x68 ) ;
  if ( the_pll.read_loss_of_signal( 0 ) == true ) return 0 ;
  return 1 ;
}

/// Frequency status of PLL (good = 1 , out of frequency = 0 )
int pcie40_pllFrequencyStatus( int dev , int pll ) {
  if ( pll > 3 ) {
    std::cout << "Invalid PLL number" << std::endl ;
    return 0 ;
  }
  pcie40_pll the_pll( dev , pll , 0x68 ) ;
  if ( the_pll.read_out_of_frequency( 0 ) == true ) return 0 ;
  return 1 ;
}

/// Loss of Lock counter (1 if no LOL detected, 0 otherwise)
int pcie40_pllLockCounter( int dev , int pll ) {
  if ( pll > 3 ) {
    std::cout << "Invalid PLL number" << std::endl ;
    return 0 ;
  }
  pcie40_pll the_pll( dev , pll , 0x68 ) ;
  if ( the_pll.read_loss_of_lock_flg( ) == true ) return 0 ;
  return 1 ;
}

/// Loss of Signal counter (1 if no LOS detected, 0 otherwise)
int pcie40_pllInputCounter( int dev , int pll ) { 
  if ( pll > 3 ) {
    std::cout << "Invalid PLL number" << std::endl ;
    return 0 ;
  }
  pcie40_pll the_pll( dev , pll , 0x68 ) ;
  if ( the_pll.read_loss_of_signal_flg( 0 ) == true ) return 0 ;
  return 1 ;
}

/// Out of Frequency counter (1 if no OOF detected, 0 otherwise)
int pcie40_pllFrequencyCounter( int dev , int pll ) {
  if ( pll > 3 ) {
    std::cout << "Invalid PLL number" << std::endl ;
    return 0 ;
  }
  pcie40_pll the_pll( dev , pll , 0x68 ) ;
  if ( the_pll.read_out_of_frequency_flg( 0 ) == true ) return 0 ;
  return 1 ;
}

/// Reset of Loss of Lock counter
void pcie40_pllLockReset( int dev , int pll ) {
  if ( pll > 3 ) {
    std::cout << "Invalid PLL number" << std::endl ;
    return  ;
  }
  pcie40_pll the_pll( dev , pll , 0x68 ) ;
  the_pll.clear_lol_flg_bit() ;
}

/// Reset of Loss of Signal counter
void pcie40_pllInputReset( int dev , int pll ) {
  if ( pll > 3 ) {
    std::cout << "Invalid PLL number" << std::endl ;
    return  ;
  }
  pcie40_pll the_pll( dev , pll , 0x68 ) ;
  the_pll.clear_los_oof_flg_reg() ;
}

/// Reset of Out of Frequency counter
void pcie40_pllFrequencyReset( int dev , int pll ) {
  if ( pll > 3 ) {
    std::cout << "Invalid PLL number" << std::endl ;
    return ;
  }
  pcie40_pll the_pll( dev , pll , 0x68 ) ;
  the_pll.clear_los_oof_flg_reg() ;
}
